/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * \file move.h
 * File contains functionality about moves.
 *
 * Move data structure:<br>
 * 1111000000000000 - type<br>
 * 0000111111000000 - from<br>
 * 0000000000111111 - to<br>
 */

#ifndef MOVE_H
#define MOVE_H

#include <glib.h>
#include "piece.h"
#include "square.h"

#ifndef Empty
#define Empty 0
#endif

#define Normal 1
#define Capture 2
#define En_Passant 3
#define Castle 4
#define Promote_Knight 5
#define Promote_Bishop 7
#define Promote_Rook 9
#define Promote_Queen 11
#define Null_Move 12

#define moveGenerate(type,from,to) (type|from<<4|(to<<10))
#define moveGetFrom(move) (move>>4&63)
#define moveGetTo(move) (move>>10&63)
#define moveGetType(move) (move&15)

#define Max_Moves 150
#define Max_Pv 14

#define START 0
#define STATIC 1
#define DINAMIC 2

typedef guint Move;

struct tMoves{
	guint size;
	guint toSort;
	guint next;
	Move move[Max_Moves];
	gint eval[Max_Moves];
	guint state;
};
typedef struct tMoves Moves;

struct tRootMove{
	Move move;
	gint eval;
	guint depth;
	guint seldepth;
	guint nps;
	guint time;
	guint64 nodes;
	guint hash;
	Move pv[Max_Pv];
};
typedef struct tRootMove RootMove;

struct tRootMoves{
	guint size;
	RootMove move[Max_Moves];
};
typedef struct tRootMoves RootMoves;

void moveToString(const Move move, gchar string[6]);
void movePrint(const Move move);
void movesPrint(const Moves* moves);
void rootMovesPrint(const RootMoves* moves);

#endif
