/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * \file position.h
 */

#ifndef POSITION_H
#define POSITION_H

#include <string.h>

#include "move.h"
#include "init.h"
#include "bit.h"

#define Max_Ply 500	

#define WHITE_CASTLE 3
#define WHITE_KING_CASTLE 1
#define WHITE_QUEEN_CASTLE 2
#define BLACK_CASTLE 12
#define BLACK_KING_CASTLE 4
#define BLACK_QUEEN_CASTLE 8

#define Opening 0
#define Endgame 1

#define positionCopy(from,to) memcpy(to, from, sizeof(Position))

struct tHistory{
	guint64 key;
	guint 	capture;
	guint 	staticMoves;
	guint 	enPassant;
	guint 	castle;
	gboolean	inCheck;
	gint    material[2];
	gint	pvt[2];
	gboolean	extend;
	gint	phase;
	Move 	move;
};
typedef struct tHistory History;

struct tPosition{
	guint64 piece[14];
	guint64 rotateL90;
	guint64 rotateL45;
	guint64 rotateR45;
	guint   type[64];
	guint   turn;
	guint   ply;
	guint64 key;
	guint	capture;
	guint   staticMoves;
	guint   enPassant;
	guint   castle;
	gboolean	inCheck;
	gint    material[2];
	gint	pvt[2];
	gboolean extend;
	gint phase;
	History history[Max_Ply];
};
typedef struct tPosition Position;

Position* positionNew();
void positionFree(Position* position);
void positionStartPosition(Position* position);
void positionPrint(const Position* position);
guint positionRepeat(const Position* position);
void positionSetFen(Position* position, const gchar fen[]);
gboolean positionIsInCheck(const Position* position);
void positionIsOk(const Position* position);

static inline guint64 positionAttackR45(const guint square, const gint64 rotateR45){
	return attackBishopR45[square][((rotateR45)>>shiftR45[square])&63];
}

static inline guint64 positionAttackL45(const guint square, const guint64 rotateL45){
	return attackBishopL45[square][((rotateL45)>>shiftL45[square])&63];
}

static inline guint64 positionAttackRank(const guint square, const guint64 pieces){
	return attackRookL00[(square)&7][((pieces)>>(((square)&~7)+1))&63]<<((square)&~7);
}

static inline guint64 positionAttackFile(const guint square, const guint64 rotateL90){
	return attackRookL90[square][((rotateL90)>>((((square)&7)<<3)+1))&63];
}

static inline guint64 positionAttackRook(const Position* position, const guint square){
	return positionAttackRank(square,position->piece[White]|position->piece[Black])|
			positionAttackFile(square,position->rotateL90);
}

static inline guint64 positionAttackBishop(const Position* position, const guint square){
	return positionAttackR45(square,position->rotateR45)|positionAttackL45(square,position->rotateL45);
}

static inline guint64 positionAttackQueen(const Position* position, const guint square){
	return positionAttackRook(position, square) | positionAttackBishop(position, square);
}

static inline gboolean positionEndGame(Position* position){
	if(numberOfBits64(position->rotateL90) <= 10 && position->phase <= 7) return TRUE;
	return FALSE;
}

static inline gint positionMaterial(Position* position){
	return position->material[position->turn] - position->material[position->turn^1];
}

gint moveMaterial(const Position* position, const Move move);

static inline gboolean positionLastMoveDangerous(const Position* position){
    Move move = position->history[position->ply-1].move;
    gint type = moveGetType(move), pieceType,square;
    if(type > Normal && type != Castle && type != Null_Move) return TRUE;
    if(type == Normal){
        square = moveGetFrom(move);
        pieceType = position->type[square];
        if(pieceType == White_Pawn && rank(square) >= Rank_6) return TRUE;
        if(pieceType == Black_Pawn && rank(square) <= Rank_3) return TRUE;
    }
    return FALSE;
}

#endif
