/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <glib/gprintf.h>

#include "position.h"
#include "search.h"
#include "threadSearch.h"
#include "position.h"
#include "moveGenerator.h"
#include "evaluator.h"
#include "moveDo.h"
#include "uci.h"
#include "hash.h"
#include "bit.h"
#include "book.h"
#include "evaluator.h"
#include "egbb.h"

gboolean stopFlag = FALSE;
gboolean parallelStopFlag = FALSE;
gboolean ponderSearch = FALSE;
gint searchMultipv = 1;
guint64 searchNodes;
guint history[64][64];
static guint searchDepth;

const gint futilityValue1 = 200;
const gint futilityMargin1 = 200;
const gint futilityValue2 = 500;
const gint futilityMargin2 = 500;
const gint futilityValue3 = 900;
const gint futilityMargin3 = 500;

const gint HistoryFullDepthMoves = 4;
const gint HistoryReductionLimit = 3;

void rootSearch(Position* position, RootMoves* moves, guint depth);
guint updateMultipv(RootMoves* moves, guint current, gint value, guint depth);


void iterativeDeepening(const Position* position, const guint64 maxNodes, const guint maxDepth){
	guint depth	= 1,i,j;
	Position searchPosition;
	Move bMove = Empty;

	searchNodes = 0;
	searchDepth = 0;
	bbhits = 0;
	
	positionCopy(position,&searchPosition);

	if(maxDepth == 0){
		   evaluate(&searchPosition, -EVAL_MATE, EVAL_MATE);
		   g_printf("bestmove xxxx\n");
		   fflush(stdout);
		   return;
	}

	RootMoves moves;
	rootMovesGenerate(&searchPosition, &moves);

	if(positionRepeat(&searchPosition) == 2 || searchPosition.staticMoves >= 100 || moves.size == 0){
		g_printf("info string in given position can not make any move!\n");
		return;
	}
	
	if(moves.size == 1){
		uciBestMove(moves.move[0].move, Empty);
		return;
	}
	
	hashIncrement();

	for(i=0; i<64; i++)
		for(j=0; j<64; j++)
			history[i][j] = 0;

	searchMultipv = MIN(searchMultipv,moves.size);

	if(isBookLoaded()){
		bMove = bookMove(&searchPosition);
		if(bMove != Empty){
			uciBestMove(bMove,Empty);
			return;
		}
	}

	while(depth <= maxDepth && searchNodes < maxNodes && (!stopFlag || ponderSearch)){
		uciBeginDepthInfo(depth);
		rootSearch(&searchPosition,&moves, depth);
		uciEndDepthInfo(depth, searchNodes);
		depth++;
	}

	uciBestMove(moves.move[0].move,moves.move[0].pv[0]);
}

void rootSearch(Position* position, RootMoves* moves, guint depth){
	gint value, alpha, beta;
	guint i, current = 0, window;
	gboolean up, down;
	HashData* hd;
	guint64 key = position->key, hData;
	
	hashFind(&hd,key,&hData);
			
	while(!stopFlag || ponderSearch){
		for(i=0; i<moves->size; i++){
			if(moves->move[i].depth < depth){
				current = i;
				break;
			}
		}
		if(i == moves->size) break;
		
		if(!moveDo(position,moves->move[current].move)) g_error("Wrong root move!\n");
		uciCurrentMove(depth,current,moves);
		
		if(current < searchMultipv){
			value = moves->move[current].eval;
			alpha = value - 1;
		}
		else{
			value = moves->move[searchMultipv-1].eval;
 			alpha = value;
		}

		window = 2;
		up = down = FALSE;
		beta = value + 1;

		while(TRUE){
			if(searchThreads == 1) value = -search(position,-beta, -alpha, depth-1, 0, TRUE);
			else value = -threadSearch(position, -beta, -alpha, depth-1, 0, TRUE);
			if(stopFlag == TRUE && !ponderSearch) break;
			if(current < searchMultipv){
				current = updateMultipv(moves, current, value, depth);
				hashCollectPV(position, moves->move[current].pv);
				uciMultiPv(position,depth,searchMultipv,moves);
				if(value >= beta){
					up = TRUE;
					if(up && down) break;
					window += window;
					alpha = value;
					beta = value + window;
					if(beta > EVAL_MATE) beta = EVAL_MATE+1;
					hashWrite(hd,key,depth,value,ALPHA,moves->move[current].move);	
					continue;
				}
				if(value <= alpha){
					down = TRUE;
					if(up && down) break;
					window += window;
					beta = value;
					alpha = value - window;
					if(alpha < -EVAL_MATE) alpha = -EVAL_MATE - 1;
					hashWrite(hd,key,depth,value,BETA,moves->move[current].move);
					continue;
				}
				hashWrite(hd,key,depth,value,EXACT,moves->move[current].move);
				break;
			}
			else{
				if(value <= alpha){
					moves->move[current].eval = value;
					moves->move[current].depth = depth;
					hashWrite(hd,key,depth,value,BETA,moves->move[current].move);
					break;
				}
				else{
					current = updateMultipv(moves, current, value, depth);
					uciMultiPv(position,depth,searchMultipv,moves);
					up = TRUE;
					window = 4;
					alpha = value;
					beta = value + window;
					if(beta > EVAL_MATE) beta = EVAL_MATE+1;
					hashWrite(hd,key,depth,value,ALPHA,moves->move[current].move);
					continue;
				}
			}
		}
		moveUndo(position);
	}
	hashWriteBestMove(position->key,depth,moves->move[0].eval,moves->move[0].move);
}

guint updateMultipv(RootMoves* moves, guint current, gint value, guint depth){
	gdouble sec = uciGetTime();
	RootMove move;
	guint i,nps;

	if(sec > 0) nps = (guint)(searchNodes/sec);
	else nps = 0;

	moves->move[current].eval = value;
	moves->move[current].depth = depth;
	moves->move[current].seldepth = searchDepth;
	moves->move[current].time = (guint)(sec*1000);
	moves->move[current].nps = nps;
	moves->move[current].nodes = searchNodes;
	for(i=0; i<current; i++){
		if(moves->move[i].eval < value){
			memcpy(&move,&moves->move[current],sizeof(RootMove));
			while(i < current){
				memcpy(&moves->move[current],&moves->move[current-1],sizeof(RootMove));
				current --;
			}
			memcpy(&moves->move[i],&move,sizeof(RootMove));
			current = i;
			break;
		}
	}
	return current;
}

gint search(Position* position, gint alpha, gint beta, gint depth, guint selDepth, gboolean nullMove){
	Moves moves;
	Move move, bestMove = Empty, hMove = Empty;
	gboolean extend=FALSE, endGamePhase, fprune = FALSE;
	gint value, tmpDepth, fmargin = 0, movesSearched = 0;
	guint64 key = position->key, hData;
	HashData* hd;

	if((stopFlag == TRUE && !ponderSearch) || parallelStopFlag == TRUE ) return EVAL_DRAW;

	searchNodes++;

	/* Draw */
	if(positionRepeat(position) >= 1 || position->staticMoves >= 100) return EVAL_DRAW;

	/* Hash */
	if(hashFind(&hd,key,&hData)){
		hMove = hashMove(hData);
        if(hashDepth(hData) >= depth){
			if(hashFlag(hData) == EXACT) return hashEval(hData);
			else if(hashFlag(hData) == ALPHA && hashEval(hData) > alpha){
				bestMove = hashMove(hData);
				alpha = hashEval(hData);
                if(alpha >= beta) return alpha;
			}
			else if(hashFlag(hData) == BETA && hashEval(hData) < beta){
				beta = hashEval(hData);
				if(alpha >= beta) return beta;
			}
		}
	}

	/* Extension */	
	if(position->extend){
		depth ++;
		extend = TRUE;
	}

	/* Horizon */
	if(depth < 1)return quiescent(position,alpha,beta,selDepth+1);

	endGamePhase = positionEndGame(position);

	/* Null move pruning */
	if (nullMove && depth>2 && !extend && !endGamePhase){
		if(depth > 6) tmpDepth = depth - 4;
		else tmpDepth = depth - 3;
		moveDo(position,Null_Move);
		value = -search(position,-beta,-beta+1,tmpDepth,selDepth+1,FALSE);
		moveUndo(position);
		if(value >= beta){
			hashWrite(hd,key,depth,value,ALPHA,Empty);
			return value;
		}
	}

	/* Internal Iterative Deepening */
	tmpDepth = depth/2;
	if(hMove == Empty && depth > 3 && tmpDepth >= hashDepth(hData)){
		search(position,alpha, beta, tmpDepth, selDepth,FALSE);
		if(hashFind(&hd,key,&hData)) hMove = hashMove(hData);
	}

	/* Futility pruning */
	if(depth <= 3 && !extend && !endGamePhase){
		value = positionMaterial(position);
		switch(depth){
			case 3:{
				if(value + futilityValue3 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin3;
					depth --;
					break;
				}
			}
			case 2:{
				if(value + futilityValue2 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin2;
					break;
				}
			}
			case 1:{
				if(value + futilityValue1 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin1;
					break;
				}
			}
		}
	}

	/* All moves */
	moves.state = START;
	if(hMove != Empty){
		move = hMove;
		moveDo(position,move);
	}
	else move = nextMoveDo(position,&moves);
	while(move != Empty){
		/* Futility pruning */
		if(fprune && !position->extend && fmargin - positionMaterial(position) <= alpha){
			moveUndo(position);
			move = nextMoveDo(position,&moves);
			movesSearched ++;
			continue;
		}
		if(movesSearched == 0){
			value = -search(position, -beta, -alpha, depth-1, selDepth+1,TRUE);
		}else{
			/* History pruning */
			if(movesSearched >= HistoryFullDepthMoves && depth > HistoryReductionLimit && 
				!extend && !position->extend){
				value = -search(position, -(alpha+1), -alpha, depth-2, selDepth+1,TRUE);
			}
			else value = alpha + 1;

			/* PVS */
			if(value > alpha){
				value = -search(position, -(alpha+1), -alpha, depth-1, selDepth+1,TRUE);
				if(value > alpha && value < beta)
					value = -search(position, -beta, -alpha, depth-1, selDepth+1,TRUE);
			}
		}
		moveUndo(position);
		if(value > alpha){
			if(value >= beta){
				if(moves.state == STATIC) history[moveGetFrom(move)][moveGetTo(move)] += 1<<depth;
				hashWrite(hd,key,depth,value,ALPHA,move);
				return value;
			}
			alpha = value;
			bestMove = move;
		}
		movesSearched ++;
		move = nextMoveDo(position,&moves);
	}

	/* Mate or pat */
	if(movesSearched == 0){
		if(position->inCheck == TRUE){
			value = -EVAL_MATE+selDepth;
			hashWrite(hd,key,depth,value,EXACT,Empty);
			return value;
		}
		else{  
			hashWrite(hd,key,depth,EVAL_DRAW,EXACT,Empty);
			return EVAL_DRAW;
		}
	}

	/* hash is update */
	if(bestMove != Empty) hashWrite(hd,key,depth,alpha,EXACT,bestMove);
	else hashWrite(hd,key,depth,alpha,BETA,hMove);

	return alpha;
}

gint quiescent(Position* position, gint alpha, gint beta, guint selDepth){
	Moves moves;
	Move move;
	gint value;	

	if((stopFlag == TRUE && !ponderSearch) || parallelStopFlag == TRUE) return -EVAL_MATE;

	searchNodes++;

	if(selDepth > searchDepth) searchDepth = selDepth;

	value = evaluate(position, alpha, beta);

	if(value > alpha)
	{
		if(value >= beta) return value;
		alpha = value;
	}

	moves.state = START;
	move = nextDinamicMoveDo(position,&moves);
	while(move != Empty){
		value = -quiescent(position, -beta, -alpha, selDepth+1);
		moveUndo(position);
		if(value > alpha){
			if(value >= beta) return value;
			alpha = value;
		}
		move = nextDinamicMoveDo(position,&moves);
	}
	return alpha;
}

